/*
    The DataBox Libary - developerfriendly working with bytes
    Copyright (C) 2000-2003  Christian Leber

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    e-mail: christian@leber.de
*/

#include <stdio.h>
#include <string.h>
#include "databox_core.h"
#include "malloc/plibs_malloc.h"


/* INIT */
/***********************************************************/
/* 
   inititialize a DataBox structure with a reserved size of
   size and return a pointer to it

   There is nothing to fail, besides plibs_malloc, plibs_malloc
   will not return if malloc fails.
*/
DataBox *
DataBox_init(int size)
{
	DataBox *output;
	
	output = (DataBox *)plibs_malloc(sizeof(DataBox));
	if(size>0)
	{
		output->storage = (unsigned char *)plibs_malloc(size);
		output->reserved=size;
	}
	else
	{
		output->reserved=0;
	}

	/* because there is no stuff yet we set the length to 0 */
	output->length=0;
	/* and data to NULL */
	output->data=NULL;

	/* now the DataBox is inited, so we set init to true */
	output->init=1;
	return output;
}
/***********************************************************/


/* DESTROY */
/***********************************************************/
/* 
   destroy the DataBox given by *input

   there is no return value
*/
void
DataBox_destroy(DataBox *input)
{
	/* if INPUT==NULL than there is nothing to free */
	if(input==NULL) return;
	
	/* if storage is >0 then there is memory allocated
	   and we have to free storage */
	if(input->reserved > 0) plibs_free(input->storage);
	/* free the DataBox data struct */
	plibs_free(input);
}
/***********************************************************/


/* SAVE RESIZE */
/***********************************************************/
/* 
   This will ensure that there is space for new_size bytes
   in the *this DataBox, it will return an error if new_size
   is smaller then the needed amount of bytes at the moment.

   It will also shrink the memory area of the DataBox if this
   is wished.

   the content of *this will not be changed 
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed
   -2 = new_size choosen to small
*/
int
DataBox_resize(DataBox *this, DATABOX_SIZE_TYPE new_size)
{
	unsigned char *new_storage;

	/* ensure that *this is usable */
	if(this == NULL) return -1;
	if(this->init != 1) return -1;

	/* ok, nothing to do */
	if(new_size == this->reserved) return 0;
	/* we are not going to destroy data */
	if(new_size < this->length) return -2;

	new_storage = (unsigned char *)plibs_malloc(new_size);

	memcpy(new_storage,this->data,this->length);

	plibs_free(this->storage);
	this->storage = new_storage;
	this->reserved = new_size;
	this->data = this->storage;

	return 0;
}
/***********************************************************/


/* CARELESS ENSURE SIZE */
/***********************************************************/
/* 
   ensures that there are size bytes space in the DataBox,
   it DONT care about the content it will be destroyed

   *this will be overwritten/reseted
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_careless_ensure_size(DataBox *this, DATABOX_SIZE_TYPE size)
{
	
	/* ensure that *this is usable */
	if(this == NULL) return -1;
	if(this->init != 1) return -1;
	
	/* check if there is allready enough space */
	if(this->reserved < size)
	{
		/* free if there is allready space allocated */
		if(this->reserved > 0)
		{
			plibs_free(this->storage);
		}
		this->storage = (unsigned char *)plibs_malloc(size);
		this->reserved = size;
	}

	/* the box is now empty */
	this->data = this->storage;
	this->length=0;
	
	return 0;
}
/***********************************************************/


/* MOVE */
/***********************************************************/
/* 
   move the content of *from to *aim and free *from

   *aim will be overwritten
   *from will be destroyed
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_move(DataBox *aim, DataBox *from)
{
	/* check if this are correct DataBoxes */
	if(aim == NULL) return -1;
	if(aim->init != 1) return -1;

	if(from == NULL) return -1;
	if(from->init != 1) return -1;

	/* if there is memory allocated in *aim, then free it */
	/* otherwise it would be lost in the nowhere... */
	if(aim->reserved > 0) plibs_free(aim->storage);
	
	/* now transfer the stuff from *from to *aim */
	aim->storage = from->storage;
	aim->data = from->data;
	aim->reserved = from->reserved;
	aim->length = from->length;
	
	/* and free *from */
	plibs_free(from);
	
	/* everything seems to be ok, so we return 0 */
	return 0;
}
/***********************************************************/


/* COPY */
/***********************************************************/
/* 
   copy the content of *from to *aim

   *aim will be overwritten
   *from stays untouched
   
   if in *from the amount on allocated memory is bigger then
   the relly needed amount of memory this will be corrected
   in aim, but this is transparent for the user of the libary
   if it is used correct.
      
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_copy(DataBox *aim, DataBox *from)
{
	/* check if this are correct DataBoxes */
	if(aim == NULL) return -1;
	if(aim->init != 1) return -1;

	if(from == NULL) return -1;
	if(from->init != 1) return -1;

	/* allocate the needed amount of memory
	   (sometimes there is more memory allocated then
	   needed, we will correct this of course in *aim) */
	aim->storage = (unsigned char *)plibs_malloc(from->length);
	
	/* set data */
	aim->data = aim->storage;

	/* now the length and the reserved amount of memory are the
	   same */
	aim->length = from->length;
	aim->reserved = from->length;
	
	/* copy the actuel data */
	memcpy(aim->data,from->data,from->length);
	
	/* everything seems to be ok, so we return 0 */
	return 0;
}
/***********************************************************/


/* APPEND */  /* FIXME works, but sucks (shoud check if there is allready enough space) */
/***********************************************************/
/* 
   append the content of *from to *aim

   *aim will be changed (of course)
   *from stays untouched
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_append(DataBox *aim, DataBox *from)
{
	int newlength; /* new length in *aim */
	int newreserved; /* new reserved size for *aim */
   	
	/* check if this are correct DataBoxes */
	if(aim == NULL) return -1;
	if(aim->init != 1) return -1;

	if(from == NULL) return -1;
	if(from->init != 1) return -1;

	/* that will be the new length of the data block */	
	newlength = aim->length + from->length;
	
	/* so much storage is needed */
	newreserved = aim->reserved + from->length;
	
	/* realloc the needed amount of memory */

	aim->storage = (unsigned char *)plibs_realloc(aim->storage,newreserved);
	
	/* copy the actual data from  *from to *aim at the right position */
	memcpy(aim->data+aim->length,from->data,from->length);
		
	/* change the length and reserved setting in *aim */
	aim->length = newlength;
	aim->reserved = newreserved;
	
	/* everything seems to be ok, so we return 0 */
	return 0;
}
/***********************************************************/


/* APPEND A SINGLE UNSIGNED CHAR */
/***********************************************************/
/* 
   append the unsigned char input to the DataBox *aim

   *aim will be changed (of course)
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_append_single_uchar(DataBox *aim, unsigned char input)
{
	int is_free_end; /* space at the end of the DB */
   	
	/* check if this are correct DataBoxes */
	if(aim == NULL) return -1;
	if(aim->init != 1) return -1;

	/* ok, now we have to check if there is enough space */
	is_free_end = DataBox_isfreeEnd(aim);
	if(is_free_end < 0) return -1; /* error */
	if(is_free_end == 0)
	{
		/* we need more space */
		/* so the next time we save time */
		if(DataBox_resize(aim,aim->length+4) < 0) return -1;
	}
	/* else (is_free_end > 0) there should be enough space */

	/* ok, so let us add the byte */
	aim->data[aim->length] = input;
	aim->length++;
	
	/* everything seems to be ok, so we return 0 */
	return 0;
}
/***********************************************************/


/* INSERT ONE BYTE AT THE BEGINNING */
/***********************************************************/
/* 
   insert one byte at the beginning of the data
   (before all other stuff)

   *workwith will be changed (of course)
   input stays untouched
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_insert_first_uchar(DataBox *workwith, unsigned char input)
{
	int newlength;
	unsigned char *newstorage;

	/* check if the DataBox is initialized */
	if(workwith==NULL) return -1;
	if(workwith->init != 1) return -1;
	
	/* the new length will be */
	newlength = workwith->length + 1;
	
	/* perhaps there is space allocated before *data */
	if(workwith->storage < workwith->data)
	{
		/* subbacool, we have not to move the memory around */
		/* so we just lower *data one byte */
		workwith->data-=1;
		workwith->length = newlength;
	}
	else
	{
		/* no, so we have to get space */
		newstorage = (unsigned char *)plibs_malloc(newlength);
		/* copy our old stuff and let a byte at the beginning free */
		memcpy(newstorage+1, workwith->data, workwith->length);
		/* free the old data area */
		plibs_free(workwith->storage);
		/* and set the variables in the structure */
		workwith->storage = newstorage;		
		workwith->data = newstorage;
		workwith->reserved = newlength;
		workwith->length = newlength;
	}
		
	
	/* so now the first byte is "free" to be overwritten */
	workwith->data[0] = input;
	/* so there is now one more byte */

	return 0;
}
/***********************************************************/


/* DELETE ONE BYTE AT THE BEGINNING */
/***********************************************************/
/* 
   delete first byte of *workwith

   *workwith will be changed
   
   will return 1 if the DataBox is empty
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed
*/
int
DataBox_delete_first_uchar(DataBox *workwith)
{
	/* check if the DataBox is initialized */
	if(workwith==NULL) return -1;
	if(workwith->init != 1) return -1;
	
	/* look if it is empty */	
	if(workwith->length==0) return -1;

	/* increase *data one byte */
	workwith->data++;
	/* so there is one byte less */
	workwith->length = workwith->length - 1;

	return 0;
}
/***********************************************************/


/* HOW MUCH SPACE IT AT THE END OF THE DATABOX ? */
/***********************************************************/
/* 
   will return the amount of free bytes at the end of the
   *workwith DataBox

   Returnvalue:
   >= 0 = all went ok, this is the amount of free memory at the end
   -1   = error, nothing changed
*/
int
DataBox_isfreeEnd(DataBox *workwith)
{
	int isfree;
	
	/* check if the DataBox is initialized */
	if(workwith==NULL) return -1;
	if(workwith->init != 1) return -1;

	if(workwith->reserved == 0 || workwith->storage == NULL)
		return 0;
	if(workwith->length == 0 || workwith->data == NULL)
		return workwith->reserved;
	
	/* this is free at all */	
	isfree = workwith->reserved - workwith->length;

	/* sub the amount of free space at the beginning */
	isfree = isfree - (workwith->data - workwith->storage); 
	
	return isfree;
}
/***********************************************************/

/* FIXME FIXME up to here all seems to be ok */

/* DATABOX TO UNSIGNED *CHAR */
/***********************************************************/
/* 
   returns an pointer to an unsigned char
   
   *input will stay untouched
   
   a memory area for unsigned chars will be malloced, and
   the content of the DataBox will copied to this area
   a pointer to this memory area will be returned

   the memory area end is marked by a 0 byte
   but if there are 0 bytes inside the DataBox they of course
   stay there.

   so it is at the moment the same as you would directy use
   *DataBox->data

   Returnvalue:
   if the DataBox is empty it will return NULL
   if there is a problem, it will return NULL
*/
unsigned char *
DataBox_DataBox_to_ucharstr(DataBox *input)
{
	unsigned char *ret_val;

	/* check if the DataBox is initialized and not empty */
	if(input == NULL) return NULL;
	if(input->init != 1) return NULL;
	if(input->length <= 0) return NULL;

	ret_val=(unsigned char *)plibs_malloc(input->length+1);
	
	memcpy(ret_val,input->data,input->length);
	ret_val[input->length] = 0;

	return ret_val;
}
/***********************************************************/


/* DATABOX TO *CHAR */
/***********************************************************/
/* 
   returns an pointer to an char
   
   *input will stay untouched
   
   a memory area with for chars will be malloced, and
   the content of the DataBox will copied to this area
   a pointer to this memory area will be returned

   the memory area end is marked by a 0 byte

   but if there are 0 bytes inside the DataBox,
   NULL will be returned
   
   so this should be a "real" *char string

   Returnvalue:
   if the DataBox is empty it will return NULL
   if there is a 0 byte in the DataBox it will return NULL
   if there is a problem, it will return NULL
*/
char *
DataBox_DataBox_to_charstr(DataBox *input)
{
	unsigned char *ret_val;
	DATABOX_SIZE_TYPE i;

	/* check if the DataBox is initialized and not empty */
	if(input == NULL) return NULL;
	if(input->init != 1) return NULL;
	if(input->length <= 0) return NULL;

	for(i=0; i < input->length; i++)
	{
		if(input->data[i]==0) return NULL;
	}

	ret_val=(char *)plibs_malloc(input->length+1);
	
	memcpy(ret_val,input->data,input->length);
	ret_val[input->length] = 0;

	return ret_val;
}
/***********************************************************/


/* UNSIGNED *CHAR TO DATABOX */
/***********************************************************/
/* 
   copy the content of the *input data string to
   the DataBox *output

   *output will be changed
   *input stays untouched
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_ucharstr_to_DataBox(DataBox *output, unsigned char *input)
{
	/* length of the *input string */
	DATABOX_SIZE_TYPE input_length;

	/* check if the DataBox is initialized and not empty */
	if(output == NULL) return -1;
	if(output->init != 1) return -1;
	if(input == NULL) return -1;

	input_length = strlen(input);

	/* this may not happen... */
	if(input_length<=0) return -1;
	
	/* check if there is enough space in the DataBox */
	DataBox_careless_ensure_size(output, input_length);

	/* copy the actual data from the char string to the DataBox */
	memcpy(output->data, input, input_length);
	output->length = input_length;

	return 0;
}
/***********************************************************/


/* UNSIGNED *CHAR + LENGTH TO DATABOX */
/***********************************************************/
/* 
   copy length bytes from *input to the DataBox *output

   *output will be changed
   *input stays untouched
   
   Returnvalue:
    0 = all went ok
   -1 = general error
   -2 = *input==NULL or/and length==0
*/
int
DataBox_ucharstr2_to_DataBox(DataBox *output, unsigned char *input, int length)
{
	/* check if the DataBox is initialized and not empty */
	if(output == NULL) return -1;
	if(output->init != 1) return -1;

	if(input == NULL) return -2;
	if(length <= 0) return -2;

	/* check if there is enough space in the DataBox */
	DataBox_careless_ensure_size(output, length);

	/* copy the actual data from the char string to the DataBox */
	memcpy(output->data, input, length);
	output->length = length;

	return 0;
}
/***********************************************************/


/* ENCODE DataBox */
/***********************************************************/
/* 
   encode a DataBox to an unsigned char, \000 and \032 will
   be encoded, they will not occur

   if the length of the DataBox is 0 or there is
   another Problem it will return -1.
   
   Returnvalue:
    -1 = general error
     0 = ok 
*/
int
DataBox_uchar_encode(DataBox *input)
{
	/* number of bytes needed because of the encoding */
	int extra=0;
	int i;
	/* this will be returned */
	unsigned char *output;
	int out_pos=0;
	
	/* check if the DataBox fits our needs */
	if(input == NULL) return -1;
	if(input->init != 1) return -1;
	if(input->length < 1) return -1;

	/* now we count the number of aditional needed bytes */
	for(i=0; i<input->length; i++)
	{
		if( input->data[i]==0 || input->data[i]==23 ||
			input->data[i]==32 ) extra++;
	}
	
	/* so the total needed amount of bytes is:
	 * input->length+extra+1                            */
	
	output = (unsigned char *)plibs_malloc(input->length+extra);
	
	for(i=0; i<input->length; i++)
	{
		if(input->data[i]==0)
		{
			output[out_pos++] = 23;
			output[out_pos++] = 1;
		}
		else if(input->data[i]==23)
		{
			output[out_pos++] = 23;
			output[out_pos++] = 23;
		
		}
		else if(input->data[i]==32)
		{
			output[out_pos++] = 23;
			output[out_pos++] = 2;

		}
		else
		{
			output[out_pos++] = input->data[i];
		}
	}
	output[out_pos]=0;

	plibs_free(input->storage);
	input->storage = output;
	input->data = output;
	input->reserved = input->length + extra;
	input->length = input->reserved;

	return 0;
}


/* DECODE DATABOX */
/***********************************************************/
/*
   decodes the *this DataBox
   counterpart to DataBox_uchar_encode
   
   Returnvalue:
    0 = all went ok
   -1 = general error
   -2 = something with the encoding seems to be wrong
        your databox could be trash
*/
int
DataBox_uchar_decode(DataBox *this)
{
	/* when this is set to 1, the last char was \023
	 * (our escape character) */
	int is_esc=0;
	/* position in the output DataBox */
	int out_pos=0;
	int i;
	
	if(this == NULL) return -1;
	if(this->init != 1) return -1;

	/* we don't have to get more space or something else
	 * because there is allready enough space
	 * i will be the position on the "input side"
	 * out_pos on the output this
	 * later we will adjust this->length according to out_pos */
	
	for(i=0;i<this->length;i++)
	{
		if(this->data[i]==23)
		{
			if(is_esc==1)
			{
				this->data[out_pos++]=23;
				is_esc=0;
			}
			else
			{
				is_esc=1;
			}
		}
		else if(this->data[i]==0 || this->data[0]==32)
		{	
			return -2;
		}
		else
		{
			if(is_esc == 1)
			{
				if(this->data[i]==1)
				{
					this->data[out_pos++] = 0;
					is_esc=0;
				}
				else if(this->data[i]==2)
				{
					this->data[out_pos++] = 32;
					is_esc=0;
				}
				else
					return -2;
			}
			else
			{
				this->data[out_pos++] = this->data[i];
			}
		}

	}
	/* ok the decoding is finished, if the last byte was an encoding sign, there
	 * is something wrong */
	if(is_esc == 1) return -2;
	if(out_pos < 1) return -2;

	this->length = out_pos;
	
	return 0;
}


/* ENCODE DataBox to ucharstr */
/***********************************************************/
/* 
   encode a DataBox to an unsigned char, \000 and \032 will
   be encoded, they will not occur

   *input stays untouched

   the function will return an memory area obtained by
   plibs_malloc, if the length of the DataBox is 0 or there is
   another Problem it will return NULL.
   
   Returnvalue:
    NULL = general error
   !NULL = pointer to the output 
*/
unsigned char *
DataBox_DataBox_to_encoded_ucharstr(DataBox *input)
{
	/* number of bytes needed because of the encoding */
	int extra=0;
	int i;
	/* this will be returned */
	unsigned char *output;
	int out_pos=0;
	
	/* check if the DataBox fits our needs */
	if(input == NULL) return NULL;
	if(input->init != 1) return NULL;
	if(input->length < 1) return NULL;

	/* now we count the number of aditional needed bytes */
	for(i=0; i<input->length; i++)
	{
		if( input->data[i]==0 || input->data[i]==23 ||
			input->data[i]==32 ) extra++;
	}
	
	/* so the total needed amount of bytes is:
	 * input->length+extra+1                            */
	
	output = (unsigned char *)plibs_malloc(input->length+extra+1);
	
	for(i=0; i<input->length; i++)
	{
		if(input->data[i]==0)
		{
			output[out_pos++] = 23;
			output[out_pos++] = 1;
		}
		else if(input->data[i]==23)
		{
			output[out_pos++] = 23;
			output[out_pos++] = 23;
		
		}
		else if(input->data[i]==32)
		{
			output[out_pos++] = 23;
			output[out_pos++] = 2;

		}
		else
		{
			output[out_pos++] = input->data[i];
		}
	}
	output[out_pos]=0;

	return output;
}


/* DECODE DATABOX to ucharstr */
/***********************************************************/
/*
   decodes the *input to the *output DataBox
   counterpart to DataBox_DataBox_to_encoded_ucharstr
   
   Returnvalue:
    0 = all went ok
   -1 = general error
   -2 = something with the encoding seems to be wrong
*/
int
DataBox_decode_ucharstr_to_DataBox(DataBox *output, unsigned char *input)
{
	/* when this is set to 1, the last char was \023
	 * (our escape character) */
	int is_esc=0;
	/* position in the output DataBox */
	int out_pos=0;
	int i;
	
	if(output == NULL) return -1;
	if(output->init != 1) return -1;
	/* first there have to be enough space in the DataBox */
	if(DataBox_careless_ensure_size(output,strlen(input)) < 0)
		return -1;
	
	for(i=0;i<strlen(input);i++)
	{
		if(input[i]==23)
		{
			if(is_esc==1)
			{
				output->data[out_pos++]=23;
				is_esc=0;
			}
			else
			{
				is_esc=1;
			}
		}
		else if(input[i]==0 || input[0]==32)
		{	
			return -2;
		}
		else
		{
			if(is_esc == 1)
			{
				if(input[i]==1)
				{
					output->data[out_pos++] = 0;
					is_esc=0;
				}
				else if(input[i]==2)
				{
					output->data[out_pos++] = 32;
					is_esc=0;
				}
				else
					return -2;
			}
			else
			{
				output->data[out_pos++] = input[i];
			}
		}

	}
	if(is_esc == 1) return -2;
	if(out_pos < 1) return -2;

	output->length = out_pos;
	
	return 0;
}

/* COMPARE TWO DATABOXES */
/***********************************************************/
/* 
   compares the content (not the memory addresses) of the
   2 DataBoxes *workwith1 and *workwith2
   
   *workwith1 and *workwith2 will stay untouched, of course
   
   Returnvalue:
    0 = all went ok, the DataBoxes have the same content
        the DataBoxes may also be empty
   -1 = the DataBoxes have not the same content
   -9 = error, one of the DataBoxes not initialized or so
*/
int
DataBox_cmp(DataBox *workwith1, DataBox *workwith2)
{
	/* check if the DataBoxes are initialized */
	if(workwith1==NULL) return -9;
	if(workwith1->init != 1) return -9;
	if(workwith2==NULL) return -9;
	if(workwith2->init != 1) return -9;

	/* first compare the size of the DataBoxes */
	if(workwith1->length != workwith2->length) return -1;
	
	/* if the length of the DataBoxes is 0, so they have the
	 * same content */
	if(workwith1->length == 0) return 0;
	
	/* now do the actual compare */
	if(memcmp(workwith1->data, workwith2->data, workwith1->length) != 0)
	{
		/* they are not equal */
		return -1;
	}
	/* so the have tobe equal */
	return 0;
}
/***********************************************************/


/* DUMP DATABOX HEXADECIMAL */
/***********************************************************/
/* 
   will dump the content of *input hexadecimal to stdout
  
   *input will stay untouched, of course
*/
void
DataBox_hexdump(DataBox *input)
{
	int i;
	/* check if the DataBoxes are initialized */
	if(input==NULL) return;
	if(input->init != 1) return;

	for(i=0;i<input->length;i++)
	{
		printf("%x ",input->data[i]);
	}
	
	printf("\n");
}
/***********************************************************/

