/*
    The DataBox Libary - developerfriendly working with bytes
    Copyright (C) 2000-2001  Christian Leber

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    e-mail: c_l@bigfoot.com
*/

#ifndef PLIBS_DATABOX_CORE_H_
#define PLIBS_DATABOX_CORE_H_

#define DATABOX_SIZE_TYPE int 

typedef struct
{
	/* here begins the allocated memory area */
	unsigned char *storage;
	/* this is the amount of allocated memory */
	DATABOX_SIZE_TYPE reserved;
	
	/* here begins the actual data */
	/* in the most cases data will be the same as storage
	   but perhaps sometimes it is wanted to just eliminate
	   the first byte, so it saves a lot of time to just change
	   the data pointer, but storage is needed to which memory
	   area has to be freed */
	unsigned char *data;
	/* the length of the data area */
	DATABOX_SIZE_TYPE length;

	/* this is used to determinate if this data struct is
	   allready initialized */
	/* is  1  if data struct has been initialized */
	unsigned char init;
} DataBox;


/*
FIXME
So what is to do if we enter a DataBox function?

The pointers name is PTR
-check if PTR!=NULL
-check if PTR->init==1
*/

/********************************/
/* base stuff */

/* INIT */
/***********************************************************/
/* 
   inititialize a DataBox structure with a reserved size of
   size and return a pointer to it

   There is nothing to fail, besides plibs_malloc, plibs_malloc
   will not return if malloc fails.
*/
DataBox *
DataBox_init(int size);


/* DESTROY */
/***********************************************************/
/* 
   destroy the DataBox given by *input

   there is no return value
*/
void
DataBox_destroy(DataBox *input);


/* SAVE RESIZE */
/***********************************************************/
/* 
   This will ensure that there is space for new_size bytes
   in the *this DataBox, it will return an error if new_size
   is smaller then the needed amount of bytes at the moment.

   It will also shrink the memory area of the DataBox if this
   is wished.

   the content of *this will not be changed 
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed
   -2 = new_size choosen to small
*/
int
DataBox_resize(DataBox *this, DATABOX_SIZE_TYPE new_size);


/* CARELESS ENSURE SIZE */
/***********************************************************/
/* 
   ensures that there are size bytes space in the DataBox,
   it DONT care about the content it will be destroyed

   *this will be overwritten/reseted
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_careless_ensure_size(DataBox *this, DATABOX_SIZE_TYPE size);


/* MOVE */
/***********************************************************/
/* 
   move the content of *from to *aim and free *from

   *aim will be overwritten
   *from will be destroyed
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_move(DataBox *aim, DataBox *from);


/* COPY */
/***********************************************************/
/* 
   copy the content of *from to *aim

   *aim will be overwritten
   *from stays untouched
   
   if in *from the amount on allocated memory is bigger then
   the relly needed amount of memory this will be corrected
   in aim, but this is transparent for the user of the libary
   if it is used correct.
      
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_copy(DataBox *aim, DataBox *from);


/* APPEND */
/***********************************************************/
/* 
   append the content of *from to *aim

   *aim will be changed (of course)
   *from stays untouched
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_append(DataBox *aim, DataBox *from);


/* APPEND A SINGLE UNSIGNED CHAR */
/***********************************************************/
/* 
   append the unsigned char input to the DataBox *aim

   *aim will be changed (of course)
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_append_single_uchar(DataBox *aim, unsigned char input);

	
/* INSERT ONE BYTE AT BEGINNING */
/***********************************************************/
/* 
   insert one byte at the beginning of the data
   (before all other stuff)

   *workwith will be changed (of course)
   input stays untouched
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_insert_first_uchar(DataBox *workwith, unsigned char input);


/* DELETE ONE BYTE AT THE BEGINNING */
/***********************************************************/
/* 
   delete first byte of *workwith

   *workwith will be changed
   
   will return 1 if the DataBox is empty
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed
*/
int
DataBox_delete_first_uchar(DataBox *workwith);


/* HOW MUCH SPACE IT AT THE END OF THE DATABOX ? */
/***********************************************************/
/* 
   will return the amount of free bytes at the end of the
   *workwith DataBox

   Returnvalue:
   >= 0 = all went ok, this is the amount of free memory at the end
   -1   = error, nothing changed
*/
int
DataBox_isfreeEnd(DataBox *workwith);


/* DATABOX TO UNSIGNED *CHAR */
/***********************************************************/
/* 
   returns an pointer to an unsigned char
   
   *input will stay untouched
   
   a memory area for unsigned chars will be malloced, and
   the content of the DataBox will copied to this area
   a pointer to this memory area will be returned

   the memory area end is marked by a 0 byte
   but if there are 0 bytes inside the DataBox they of course
   stay there.

   so it is at the moment the same as you would directy use
   *DataBox->data

   Returnvalue:
   if the DataBox is empty it will return NULL
   if there is a problem, it will return NULL
*/
unsigned char *
DataBox_DataBox_to_ucharstr(DataBox *input);


/* DATABOX TO *CHAR */
/***********************************************************/
/* 
   returns an pointer to an char
   
   *input will stay untouched
   
   a memory area with for chars will be malloced, and
   the content of the DataBox will copied to this area
   a pointer to this memory area will be returned

   the memory area end is marked by a 0 byte

   but if there are 0 bytes inside the DataBox,
   NULL will be returned
   
   so this should be a "real" *char string

   Returnvalue:
   if the DataBox is empty it will return NULL
   if there is a 0 byte in the DataBox it will return NULL
   if there is a problem, it will return NULL
*/
char *
DataBox_DataBox_to_charstr(DataBox *input);


/* UNSIGNED *CHAR TO DATABOX */
/***********************************************************/
/* 
   copy the content of the *input data string to
   the DataBox *output

   *output will be changed
   *input stays untouched
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_ucharstr_to_DataBox(DataBox *output, unsigned char *input);


/* UNSIGNED *CHAR + LENGTH TO DATABOX */
/***********************************************************/
/* 
   copy length bytes from *input to the DataBox *output

   *output will be changed
   *input stays untouched
   
   Returnvalue:
    0 = all went ok
   -1 = general error
   -2 = *input==NULL or/and length==0
*/
int
DataBox_ucharstr2_to_DataBox(DataBox *output, unsigned char *input, int length);


/* ENCODE DataBox */
/***********************************************************/
/* 
   encode a DataBox to an unsigned char, \000 and \032 will
   be encoded, they will not occur

   if the length of the DataBox is 0 or there is
   another Problem it will return -1.
   
   Returnvalue:
    -1 = general error
     0 = ok 
*/
int
DataBox_uchar_encode(DataBox *input);


/* DECODE DATABOX */
/***********************************************************/
/*
   decodes the *this DataBox
   counterpart to DataBox_uchar_encode
   
   Returnvalue:
    0 = all went ok
   -1 = general error
   -2 = something with the encoding seems to be wrong
        your databox could be trash
*/
int
DataBox_uchar_decode(DataBox *this);


/* ENCODE DataBox to ucharstr */
/***********************************************************/
/* 
   encode a DataBox to an unsigned char, \000 and \032 will
   be encoded, they will not occur

   *input stays untouched

   the function will return an memory area obtained by
   plibs_malloc, if the length of the DataBox is 0 or there is
   another Problem it will return NULL.
   
   Returnvalue:
    NULL = general error
   !NULL = pointer to the output 
*/
unsigned char *
DataBox_DataBox_to_encoded_ucharstr(DataBox *input);


/* DECODE DATABOX to ucharstr */
/***********************************************************/
/*
   decodes the *input to the *output DataBox
   counterpart to DataBox_DataBox_to_encoded_ucharstr

   Returnvalue:
    0 = all went ok
   -1 = general error
   -2 = something with the encoding seems to be wrong
*/
int
DataBox_decode_ucharstr_to_DataBox(DataBox *output, unsigned char *input);


/* COMPARE TWO DATABOXES */
/***********************************************************/
/* 
   compares the content (not the memory addresses) of the
   2 DataBoxes *workwith1 and *workwith2
   
   *workwith1 and *workwith2 will stay untouched, of course
   
   Returnvalue:
    0 = all went ok, the DataBoxes have the same content
        the DataBoxes may also be empty
   -1 = the DataBoxes have not the same content
   -9 = error, one of the DataBoxes not initialized or so
*/
int
DataBox_cmp(DataBox *workwith1, DataBox *workwith2);


/* DUMP DATABOX HEXADECIMAL */
/***********************************************************/
/* 
   will dump the content of *input hexadecimal to stdout
  
   *input will stay untouched, of course
*/
void
DataBox_hexdump(DataBox *input);


#endif /* PLIBS_DATABOX_CORE_H_ */
