/*
    databox_crypto_blowfish.c - simple to use blowfish crypto functions
    Copyright (C) 2000-2001  Christian Leber

    This file is part of the DataBox crypto libary.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    e-mail: christian@leber.de
*/

#include <stdlib.h> /* for div_t */
#include "databox_crypto_blowfish.h"
#include "malloc/plibs_malloc.h"
#include "random.h"


/* INIT DataBoxBLOWFISH */
/***********************************************************/
/*
   return a DataBoxBLOWFISH data structure

   Returnvalues:
   pointer to the data structure
*/
DataBoxBLOWFISH *
DataBox_blowfish_init(void)
{
	DataBoxBLOWFISH *output;

	output = (DataBoxBLOWFISH *)plibs_malloc(sizeof(*output));
	output->init = 1;
	output->blowfish_key = NULL;
	output->blowfish_length = 0;
	
	output->sched = NULL;
	output->sched_init = 0;
	
	return output;
}
/***********************************************************/


/* GENERATE DataBoxBLOWFISH KEY */
/***********************************************************/
/*
   generates a random blowfish key in the DataBoxBLOWFISH
   struct *input with the length of length bytes.

   Returnvalues:
    0 = all went ok
   -1 = keylength to big or to small, makes no sens
   -9 = *input not ok (not initialized or so)
*/
int
DataBox_blowfish_keygen(DataBoxBLOWFISH *input, int length)
{
	if(input==NULL) return -9;
	if(input->init != 1) return -9;
	
	/* check if the wanted length is valid */
	if( (length>72) || (length<10) ) return -1;
	
	/* so we are going to set a new key, but if there is
	 * allready one or even an schedule, we have to free it */
	if(input->blowfish_key != NULL)
	{
		/* ok, we have to free some stuff */
		plibs_free(input->blowfish_key);
		/* now look whats about the schedule */
		if(input->sched_init == 1)
		{
			/* ok, we have to free the schedule */
			free(input->sched);
			input->sched = NULL;
			input->sched_init = 0;
		}
	}
	
	input->blowfish_key = (unsigned char*)n_byte_random(length);
	input->blowfish_length = length;
	return 0;
}
/***********************************************************/


/* GENERATE SCHEDULE FOR A DataBoxBLOWFISH */
/***********************************************************/
/*
   generates a random blowfish key in the DataBoxBLOWFISH
   struct *input with the length of length bytes.

   Returnvalues:
    0 = all went ok
   -1 = keylength to big for the schedule
   -9 = *input not ok (not initialized or so)
*/
int
DataBox_blowfish_schedule(DataBoxBLOWFISH *input)
{
	int ret_val;
	
	if(input==NULL) return -9;
	if(input->init != 1) return -9;

	if(input->blowfish_length>72) return -1; //FIXME length still ok?
	input->sched = (BLOWFISH_context *)plibs_malloc( sizeof(BLOWFISH_context) );
	ret_val = bf_setkey(input->sched,input->blowfish_key,input->blowfish_length);
	if(ret_val<0) printf("bf_setkey error: %d",ret_val); //FIXME
	input->sched_init=1;
	return 0;
}
/***********************************************************/


/*
Returns -1 if the DataBox is empty
Returns -2 if there ist something wrong with the blowfish key
-9 = the data structures are not initialized or something
*/
int DataBox_blowfish_encrypt(DataBox *input, DataBoxBLOWFISH *input_key)
{
	div_t length8; /* number of bytes in the input divided by 8 */
	int length8_blocks; /* the number of 8 byte blocks */
	int i, j;
	unsigned char daten[8]; /* for the 8 byte chunks */
	int ipos; /* position while working through the data */
	unsigned char *temp; /* temporary usage */
	unsigned char *random;

	/* check if input stuff is NULL */
	if(input==NULL) return -9;
	if(input->init != 1) return -9;
	if(input_key==NULL) return -9;
	if(input_key->init != 1) return -9;

	/* check if the length of the input is >0 */
	if(input->length <= 0) return -2;

	/* insert one random byte at the beginning of the message */
	random = (unsigned char*)n_byte_random(1);
	if(DataBox_insert_first_uchar(input, random[0]) < 0 ) return -9;

	/* look if the blowfish schedule is allready created */
	if(input_key->sched_init != 1)
	{
		if(DataBox_blowfish_schedule(input_key)<0) return -2;
	}

	/* length8 is the number of 8 byte long parts, for blowfish */
	length8 = div(input->length, 8);
	if( length8.rem == 0)
	{
		/* now we've to find out if we had to add another block or not */
		if(input->data[input->length-1] >= 1 && input->data[input->length-1] <= 8)
		{
			/* ok, the extra block is needed */
			temp = (unsigned char *)plibs_malloc( (length8.quot+1) * 8 );
			memcpy(temp,input->data,input->length);
			for(i=input->length; i < (length8.quot+1)*8-1 ; i++)
			{
				temp[i] = 153;
			}
			temp[(length8.quot+1)*8-1] = 8;
			free(input->data);
			input->storage = temp;
			input->data = temp;
			input->length=(length8.quot+1)*8;
			input->reserved=input->length;
			length8_blocks=length8.quot+1;			
		}
		else
		{
			length8_blocks=length8.quot;			
		}		   
	}
	else
	{
		/* allocate the needed amount of memory */
		temp = (unsigned char *)plibs_malloc( (length8.quot+1) * 8 );
		memcpy(temp,input->data,input->length);
		for(i=input->length; i < (length8.quot+1)*8-1 ; i++)
		{
			temp[i] = 153;
		}
		temp[(length8.quot+1)*8-1] = length8.rem;
		free(input->data);
		input->storage = temp;
		input->data = temp;
		input->length=(length8.quot+1)*8;
		input->reserved=input->length;
		length8_blocks=length8.quot+1;
	}

	/* working through the first 8 byte blocks */

	encrypt_block(input_key->sched,&input->data[0],&input->data[0]);

	/* working through the other 8 byte blocks */

	for(j=1;j<length8_blocks;j++)
	{
		ipos = j * 8;
		for(i=0;i<8;i++)
		{
			input->data[ipos+i] = input->data[ipos+i-8] ^ input->data[ipos+i];
		}
		encrypt_block(input_key->sched,&input->data[ipos],&input->data[ipos]);
	}

	return 0;
}


int DataBox_blowfish_decrypt(DataBox *input, DataBoxBLOWFISH *input_key)
{
	int length8; /* the number of 8 byte blocks */
	int i, j;
	unsigned char daten[8]; /* for the 8 byte chunks */
	unsigned char next[8];
	unsigned char next_sich;
	int ipos; /* position while working through the data */
	int len;

	/* check if input stuff is NULL */
	if(input==NULL) return -1;
	if(input->init != 1) return -1;
	if(input_key==NULL) return -1;

	/* check if the length of the input is >0 */
	if(input->length <= 0) return -1;

	/* check if there are n * 8 byte */
	length8 = (input->length / 8);
	if( (length8 * 8) - input->length != 0) return -1;


	/* look if the blowfish schedule is allready created */
	if(input_key->sched_init != 1)
	{
		if(DataBox_blowfish_schedule(input_key)<0)
			return -2;
	}

	/* the first 8 byte block */

	for(i=0;i<8;i++)
		next[i] = input->data[i];
	decrypt_block(input_key->sched,&input->data[0],next);

	/* working through the other 8 byte blocks */

	for(j=1;j<length8;j++)
	{
		ipos = j * 8;
		decrypt_block(input_key->sched,daten,&input->data[ipos]);
		for(i=0;i<8;i++)
		{
			next_sich = input->data[ipos+i];
			input->data[ipos+i]=daten[i] ^ next[i];
			next[i] = next_sich;
		}
	}

	/* input->data contains now the decrypted data and unneeded extra bytes ;) */
	len=input->length;

	if(input->data[len-1]>=1 && input->data[len-1]<=7)
		input->length = len - 8 + input->data[len-1];
	else if(input->data[len-1]==8)
		input->length=len - 8;

	/* now remove the random byte at the beginning */
	DataBox_delete_first_uchar(input);

	return 0;
}


/* copy the key from a DS_blowfish data structure to a DataBox */
int DataBox_DataBoxBLOWFISH_to_DataBox(DataBox *output, DataBoxBLOWFISH *input)
{
	int i;

	/* check if input stuff seems to be ok */
	if(input==NULL) return -1;
	if(input->init != 1) return -1;
	if(input->blowfish_length<8) return -1; /* 64 bit is the minimum ! */
	
	if(output==NULL) return -1;
	if(output->init != 1) return -1;
	
	DataBox_careless_ensure_size(output, input->blowfish_length);
	
	for(i=0 ; i<input->blowfish_length ; i++)
	{
		output->data[i] = input->blowfish_key[i];
	}
	output->length = input->blowfish_length;
	return 0;
}


/* copy the key from a DataBox to a DS_blowfish data structure */
int DataBox_DataBox_to_DataBoxBLOWFISH(DataBoxBLOWFISH *output, DataBox *input)
{
	int i;

	/* check if input stuff seems to be ok */
	if(input==NULL) return -1;
	if(input->init != 1) return -1;
	if(input->length<8) return -1; /* 64 bit is the minimum ! */
	
	if(output==NULL) return -1;
	if(output->init != 1) return -1;

	/* check if the key would be too long */
	if(input->length>72) return -2;

	if(output->blowfish_key==NULL || output->blowfish_length <= input->length)
	{
		if(output->blowfish_key!=NULL) plibs_free(output->blowfish_key);
		output->blowfish_key = (unsigned char *)plibs_malloc(input->length);
	}

	for(i=0 ; i < input->length ; i++)
		output->blowfish_key[i] = input->data[i];
	output->blowfish_length = input->length;

	/* if there was allready a schedule, then it would not be the right one */
	output->sched_init = 0;
	
	return 0;
}

