/*
    databox_mpz.c - convert DataBox zu mpz and reverse
    Copyright (C) 2000-2001  Christian Leber

    This file is part of the DataBox crypto libary.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    e-mail: c_l@bigfoot.com
*/


#include <stdlib.h>
#include "databox_mpz.h"

/* MPZ TO DATABOX */
/***********************************************************/
/* 
   reads the mpz_t input and converts it to a DataBox *output

   input    will stay untouched
   *output  will be destroyed, and filled up with the output
   
   if input is <= 0 then an error (1) will be returned 
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_mpz_to_DataBox(DataBox *output, mpz_t input)
{
	/* number of bits needed to save the mpz */
	unsigned long length2;
	/* number of bytes needed to save the mpz */
	unsigned long length8;
	/* lenght2/8 */
	div_t erg;		
	unsigned long i;
	mpz_t rem, back;

	/* proof if input > 0 */
	if(mpz_cmp_ui(input,0)<=0) return -1;
	/* check if *output is usable */
	if(output==NULL) return -1;
	if(output->init != 1) return -1;
	
	/* get the number of bits in the mpz number */
	length2=mpz_sizeinbase(input,2);

	erg=div(length2,8);
	if(erg.rem==0) length8 = erg.quot;
		else length8 = erg.quot + 1;
		
	/* this ensures that there is enough space in the *output */
	DataBox_careless_ensure_size(output,length8);
	output->data = output->storage;
	output->length = length8;
	
	mpz_init(rem);
	mpz_init_set(back,input);

	for (i=0;i<length8;i++)
	{
		mpz_tdiv_qr_ui(back, rem, back, 256);
		output->data[i]=mpz_get_ui(rem);
	}

	mpz_clear(back);
	mpz_clear(rem);
	
	return 0;
}
/***********************************************************/


/* DATABOX TO MPZ */
/***********************************************************/
/* 
   reads the DataBox *input and converts it to a mpz_t output

   *input    will stay untouched
   output    will be changed
   
   If the length of the *input is < 1 an error will be
   returned
   
   Returnvalue:
    0 = all went ok
   -1 = error, nothing changed 
*/
int
DataBox_DataBox_to_mpz(mpz_t output, DataBox *input)
{
	int i;
	mpz_init(output);

	/* check if *input is usable */
	if(input==NULL) return -1;
	if(input->init != 1) return -1;
	/* databoxes with a length of 0 make here no sens */
	if(input->length < 1) return -1;

	for(i=input->length-1;i>=0;i--)
	{
		mpz_mul_2exp (output, output, 8);
		mpz_add_ui (output, output, input->data[i]);
	}
	return 0;
}
/***********************************************************/
